@@ -24,7 +24,6 @@
 
 #include <config.h>
 #include "eel-background.h"
-
 #include "eel-gdk-extensions.h"
 #include "eel-gdk-pixbuf-extensions.h"
 #include "eel-glib-extensions.h"
@@ -32,16 +31,42 @@
 #include "eel-gtk-macros.h"
 #include "eel-lib-self-check-functions.h"
 #include "eel-string.h"
+#include "eel-marshal.h"
+#include "eel-types.h"
+#include "eel-type-builtins.h"
 #include <gtk/gtkprivate.h>
 #include <gtk/gtkselection.h>
 #include <gtk/gtksignal.h>
 #include <libart_lgpl/art_rgb.h>
-#include <libgnomecanvas/gnome-canvas-util.h>
-#include <libgnomecanvas/gnome-canvas.h>
+#include <eel/eel-canvas.h>
+#include <eel/eel-canvas-util.h>
 #include <libgnomevfs/gnome-vfs-async-ops.h>
 #include <math.h>
 #include <stdio.h>
 
+/* FIXME: This could really be eliminated now */
+typedef struct {
+	/* 24-bit RGB buffer for rendering */
+	guchar *buf;
+
+	/* Rectangle describing the rendering area */
+	ArtIRect rect;
+
+	/* Rowstride for the buffer */
+	int buf_rowstride;
+
+	/* Background color, given as 0xrrggbb */
+	guint32 bg_color;
+
+	/* Invariant: at least one of the following flags is true. */
+
+	/* Set when the render rectangle area is the solid color bg_color */
+	unsigned int is_bg : 1;
+
+	/* Set when the render rectangle area is represented by the buf */
+	unsigned int is_buf : 1;
+} EelCanvasBuf;
+
 static void     eel_background_class_init                (gpointer       klass);
 static void     eel_background_init                      (gpointer       object,
 							  gpointer       klass);
@@ -50,12 +75,20 @@ static void     eel_background_start_loading_image       (EelBackground *backgro
 							  gboolean       emit_appearance_change,
 							  gboolean       load_async);
 static gboolean eel_background_is_image_load_in_progress (EelBackground *background);
+static void     eel_background_draw_to_canvas            (EelBackground *background,
+							  EelCanvasBuf  *buffer,
+							  int            entire_width,
+							  int            entire_height);
+static void     eel_background_draw_aa                   (EelBackground *background,
+							  EelCanvasBuf  *buffer);
+					  
 
 EEL_CLASS_BOILERPLATE (EelBackground, eel_background, GTK_TYPE_OBJECT)
 
 enum {
 	APPEARANCE_CHANGED,
 	SETTINGS_CHANGED,
+	DETERMINE_IMAGE_PLACEMENT,
 	IMAGE_LOADING_DONE,
 	RESET,
 	LAST_SIGNAL
@@ -104,6 +137,8 @@ struct EelBackgroundDetails {
 	int background_entire_height;
 	GdkColor background_color;
 	gboolean background_changes_with_size;
+
+	gboolean use_base;
 };
 
 static void
@@ -115,6 +150,8 @@ eel_background_class_init (gpointer klass)
 	object_class = G_OBJECT_CLASS (klass);
 	background_class = EEL_BACKGROUND_CLASS (klass);
 
+	eel_type_init ();
+
 	signals[APPEARANCE_CHANGED] =
 		g_signal_new ("appearance_changed",
 			      G_TYPE_FROM_CLASS (object_class),
@@ -135,6 +172,16 @@ eel_background_class_init (gpointer klass)
 			      g_cclosure_marshal_VOID__VOID,
 			      G_TYPE_NONE,
 			      0);
+	signals[DETERMINE_IMAGE_PLACEMENT] = 
+		g_signal_new ("determine_image_placement",
+			      G_TYPE_FROM_CLASS (object_class),
+			      G_SIGNAL_RUN_LAST | G_SIGNAL_NO_RECURSE,
+			      G_STRUCT_OFFSET (EelBackgroundClass,
+					       determine_image_placement),
+			      NULL, NULL,
+			      eel_marshal_ENUM__INT_INT,
+			      EEL_TYPE_BACKGROUND_IMAGE_PLACEMENT,
+			      2, G_TYPE_INT, G_TYPE_INT);
 	signals[IMAGE_LOADING_DONE] =
 		g_signal_new ("image_loading_done",
 			      G_TYPE_FROM_CLASS (object_class),
@@ -357,7 +404,7 @@ eel_background_ensure_gradient_buffered (EelBackground *background, int dest_wid
 }
 
 static void
-canvas_gradient_helper_v (const GnomeCanvasBuf *buf, const art_u8 *gradient_buff)
+canvas_gradient_helper_v (const EelCanvasBuf *buf, const art_u8 *gradient_buff)
 {
 	int width  = buf->rect.x1 - buf->rect.x0;
 	int height = buf->rect.y1 - buf->rect.y0;
@@ -377,7 +424,7 @@ canvas_gradient_helper_v (const GnomeCanvasBuf *buf, const art_u8 *gradient_buff
 }
 
 static void
-canvas_gradient_helper_h (const GnomeCanvasBuf *buf, const art_u8 *gradient_buff)
+canvas_gradient_helper_h (const EelCanvasBuf *buf, const art_u8 *gradient_buff)
 {
 	int width  = buf->rect.x1 - buf->rect.x0;
 	int height = buf->rect.y1 - buf->rect.y0;
@@ -396,7 +443,26 @@ canvas_gradient_helper_h (const GnomeCanvasBuf *buf, const art_u8 *gradient_buff
 }
 
 static void
-fill_canvas_from_gradient_buffer (const GnomeCanvasBuf *buf, const EelBackground *background)
+fill_rgb (EelCanvasBuf *buf, art_u8 r, art_u8 g, art_u8 b)
+{
+	art_u8 *dst = buf->buf;
+	int width = buf->rect.x1 - buf->rect.x0;
+	int height = buf->rect.y1 - buf->rect.y0;
+
+	if (buf->buf_rowstride == width * 3) {
+	 	art_rgb_fill_run (dst, r, g, b, width * height);
+	} else {
+		art_u8 *dst_limit = dst + height * buf->buf_rowstride;
+		while (dst < dst_limit) {
+	 		art_rgb_fill_run (dst, r, g, b, width);
+			dst += buf->buf_rowstride;
+		}
+	}
+}
+
+
+static void
+fill_canvas_from_gradient_buffer (const EelCanvasBuf *buf, const EelBackground *background)
 {
 	g_return_if_fail (background->details->gradient_buffer != NULL);
 
@@ -414,24 +480,24 @@ fill_canvas_from_gradient_buffer (const GnomeCanvasBuf *buf, const EelBackground
 	if (background->details->gradient_is_horizontal) {
 		if (buf->rect.x1 > background->details->gradient_num_pixels) {
 			art_u8 *rgb888 = background->details->gradient_buffer + (background->details->gradient_num_pixels - 1) * 3;
-			GnomeCanvasBuf gradient = *buf;
-			GnomeCanvasBuf overflow = *buf;
+			EelCanvasBuf gradient = *buf;
+			EelCanvasBuf overflow = *buf;
 			gradient.rect.x1 =  gradient.rect.x0 < background->details->gradient_num_pixels ? background->details->gradient_num_pixels : gradient.rect.x0;
 			overflow.buf += (gradient.rect.x1 - gradient.rect.x0) * 3;
 			overflow.rect.x0 = gradient.rect.x1;
-			eel_gnome_canvas_fill_rgb (&overflow, rgb888[0], rgb888[1], rgb888[2]);
+			fill_rgb (&overflow, rgb888[0], rgb888[1], rgb888[2]);
 			canvas_gradient_helper_h (&gradient, background->details->gradient_buffer);
 			return;
 		}
 	} else {
 		if (buf->rect.y1 > background->details->gradient_num_pixels) {
 			art_u8 *rgb888 = background->details->gradient_buffer + (background->details->gradient_num_pixels - 1) * 3;
-			GnomeCanvasBuf gradient = *buf;
-			GnomeCanvasBuf overflow = *buf;
+			EelCanvasBuf gradient = *buf;
+			EelCanvasBuf overflow = *buf;
 			gradient.rect.y1 = gradient.rect.y0 < background->details->gradient_num_pixels ? background->details->gradient_num_pixels : gradient.rect.y0;
 			overflow.buf += (gradient.rect.y1 - gradient.rect.y0) * gradient.buf_rowstride;
 			overflow.rect.y0 = gradient.rect.y1;
-			eel_gnome_canvas_fill_rgb (&overflow, rgb888[0], rgb888[1], rgb888[2]);
+			fill_rgb (&overflow, rgb888[0], rgb888[1], rgb888[2]);
 			canvas_gradient_helper_v (&gradient, background->details->gradient_buffer);
 			return;
 		}
@@ -443,7 +509,7 @@ fill_canvas_from_gradient_buffer (const GnomeCanvasBuf *buf, const EelBackground
 /* Initializes a pseudo-canvas buf so canvas drawing routines can be used to draw into a pixbuf.
  */
 static void
-canvas_buf_from_pixbuf (GnomeCanvasBuf* buf, GdkPixbuf *pixbuf, int x, int y, int width, int height)
+canvas_buf_from_pixbuf (EelCanvasBuf* buf, GdkPixbuf *pixbuf, int x, int y, int width, int height)
 {
 	buf->buf =  gdk_pixbuf_get_pixels (pixbuf);
 	buf->buf_rowstride =  gdk_pixbuf_get_rowstride (pixbuf);
@@ -631,7 +697,7 @@ eel_background_unrealize (EelBackground *background)
 	background->details->background_entire_height = 0;
 }
 
-static void
+static gboolean
 eel_background_ensure_realized (EelBackground *background, GdkWindow *window,
 				int entire_width, int entire_height)
 {
@@ -642,6 +708,7 @@ eel_background_ensure_realized (EelBackground *background, GdkWindow *window,
 	GdkGC *gc;
 	GtkWidget *widget;
 	GtkStyle *style;
+	gboolean changed;
 
 	/* Try to parse the color spec.  If we fail, default to the style's color */
 
@@ -655,7 +722,12 @@ eel_background_ensure_realized (EelBackground *background, GdkWindow *window,
 		g_assert (widget != NULL);
 
 		style = gtk_widget_get_style (widget);
-		background->details->background_color = style->bg[GTK_STATE_NORMAL];
+
+		if (background->details->use_base) {
+			background->details->background_color = style->base[GTK_STATE_NORMAL];
+		} else {
+			background->details->background_color = style->bg[GTK_STATE_NORMAL];
+		}
 	}
 
 	g_free (start_color_spec);
@@ -665,20 +737,21 @@ eel_background_ensure_realized (EelBackground *background, GdkWindow *window,
 	 */
 	if (background->details->background_pixmap != NULL &&
 	    !background->details->background_changes_with_size) {
-		return;
+		return FALSE;
 	}
 
 	/* If the window size is the same as last time, don't update */
 	if (entire_width == background->details->background_entire_width &&
 	    entire_height == background->details->background_entire_height) {
-		return;
+		return FALSE;
 	}
 
 	if (background->details->background_pixmap != NULL) {
 		g_object_unref (background->details->background_pixmap);
 		background->details->background_pixmap = NULL;
 	}
-	
+
+	changed = FALSE;
 	if (get_pixmap_size (background, entire_width, entire_height,
 			      &pixmap_width, &pixmap_height, &background->details->background_changes_with_size)) {
 		pixmap = gdk_pixmap_new (window, pixmap_width, pixmap_height, -1);
@@ -689,10 +762,13 @@ eel_background_ensure_realized (EelBackground *background, GdkWindow *window,
 				     pixmap_width, pixmap_height);
 		g_object_unref (gc);
 		background->details->background_pixmap = pixmap;
+		changed = TRUE;
 	}
 
 	background->details->background_entire_width = entire_width;
 	background->details->background_entire_height = entire_height;
+	
+	return changed;
 }
 
 GdkPixmap *
@@ -798,10 +874,10 @@ eel_background_draw (EelBackground *background,
 	int x_canvas, y_canvas;
 	int width, height;
 	
-	GnomeCanvasBuf buffer;
+	EelCanvasBuf buffer;
 	GdkPixbuf *pixbuf;
 
-	/* Non-aa background drawing is done by faking up a GnomeCanvasBuf
+	/* Non-aa background drawing is done by faking up a EelCanvasBuf
 	 * and passing it to the aa code.
 	 *
 	 * The width and height were chosen to match those used by gnome-canvas
@@ -861,7 +937,7 @@ eel_background_draw_to_pixbuf (EelBackground *background,
 			       int entire_width,
 			       int entire_height)
 {
-	GnomeCanvasBuf fake_buffer;
+	EelCanvasBuf fake_buffer;
 
 	g_return_if_fail (background != NULL);
 	g_return_if_fail (pixbuf != NULL);
@@ -874,9 +950,143 @@ eel_background_draw_to_pixbuf (EelBackground *background,
 				       entire_height);
 }
 
+static void
+canvas_draw_pixbuf_helper (art_u8 *dst, int dst_rowstride,
+			   const art_u8 *src, int src_rowstride,
+			   int copy_width, int copy_height)
+{
+	art_u8 *dst_limit = dst + copy_height * dst_rowstride;
+	int dst_bytes_per_row = copy_width * 3;
+	
+	while (dst < dst_limit) {
+ 		memcpy (dst, src, dst_bytes_per_row);
+		dst += dst_rowstride;
+		src += src_rowstride;
+	}
+}
+
+static void
+canvas_draw_pixbuf_helper_alpha (art_u8 *dst, int dst_rowstride,
+				 const art_u8 *src, int src_rowstride,
+				 int copy_width, int copy_height)
+{
+	art_u8 *dst_limit = dst + copy_height * dst_rowstride;
+	int dst_bytes_per_row = copy_width * 3;
+	
+	while (dst < dst_limit) {
+	
+		art_u8 *dst_p = dst;
+		art_u8 *dst_p_limit = dst + dst_bytes_per_row;
+		
+		const art_u8 *src_p = src;
+		
+		while (dst_p < dst_p_limit) {
+			int alpha = src_p[3];
+			if (alpha) {
+				if (alpha == 255) {
+					dst_p[0] = src_p[0];
+					dst_p[1] = src_p[1];
+					dst_p[2] = src_p[2];
+				} else {
+		  			int tmp;
+					art_u8 bg_r = dst_p[0];
+					art_u8 bg_g = dst_p[1];
+					art_u8 bg_b = dst_p[2];
+
+					tmp = (src_p[0] - bg_r) * alpha;
+					dst_p[0] = bg_r + ((tmp + (tmp >> 8) + 0x80) >> 8);
+					tmp = (src_p[1] - bg_g) * alpha;
+					dst_p[1] = bg_g + ((tmp + (tmp >> 8) + 0x80) >> 8);
+					tmp = (src_p[2] - bg_b) * alpha;
+					dst_p[2] = bg_b + ((tmp + (tmp >> 8) + 0x80) >> 8);		  
+				}
+			}
+			
+			dst_p += 3;
+			src_p += 4;
+		}
+		
+		dst += dst_rowstride;
+		src += src_rowstride;
+	}
+}
+
+/* Draws a pixbuf into a canvas update buffer (unscaled). The x,y coords are the location
+ * of the pixbuf in canvas space (NOT relative to the canvas buffer).
+ */
+static void
+canvas_draw_pixbuf (EelCanvasBuf *buf, const GdkPixbuf *pixbuf, int x, int y)
+{
+	art_u8 *dst;
+	int pixbuf_width, pixbuf_height;
+
+	/* copy_left/top/right/bottom define the rect of the pixbuf (pixbuf relative)
+	 * we will copy into the canvas buffer
+	 */
+	int copy_left, copy_top, copy_right, copy_bottom;
+	
+	dst = buf->buf;
+
+	pixbuf_width = gdk_pixbuf_get_width (pixbuf);
+	pixbuf_height = gdk_pixbuf_get_height (pixbuf);
+
+	if (x > buf->rect.x0) {
+		copy_left = 0;
+		dst += (x - buf->rect.x0) * 3;
+	} else {
+		copy_left = buf->rect.x0 - x;
+	}
+	
+	if (x + pixbuf_width > buf->rect.x1) {
+		copy_right = buf->rect.x1 - x;
+	} else {
+		copy_right = pixbuf_width;		
+	}
+	
+	if (copy_left >= copy_right) {
+		return;
+	}
+	
+	if (y > buf->rect.y0) {
+		dst += (y - buf->rect.y0) * buf->buf_rowstride;
+		copy_top = 0;
+	} else {
+		copy_top = buf->rect.y0 - y;
+	}
+	
+	if (y + pixbuf_height > buf->rect.y1) {
+		copy_bottom = buf->rect.y1 - y;
+	} else {
+		copy_bottom = pixbuf_height;		
+	}
+
+	if (copy_top >= copy_bottom) {
+		return;
+	}
+
+	if (gdk_pixbuf_get_has_alpha (pixbuf)) {
+		canvas_draw_pixbuf_helper_alpha (
+			dst,
+			buf->buf_rowstride,
+			gdk_pixbuf_get_pixels (pixbuf) + copy_left * 4 + copy_top * gdk_pixbuf_get_rowstride (pixbuf),
+			gdk_pixbuf_get_rowstride (pixbuf),
+			copy_right - copy_left,
+			copy_bottom - copy_top);
+	} else {
+		canvas_draw_pixbuf_helper (
+			dst,
+			buf->buf_rowstride,
+			gdk_pixbuf_get_pixels (pixbuf) + copy_left * 3 + copy_top * gdk_pixbuf_get_rowstride (pixbuf),
+			gdk_pixbuf_get_rowstride (pixbuf),
+			copy_right - copy_left,
+			copy_bottom - copy_top);
+	}
+}
+
+
 /* fill the canvas buffer with a tiled pixbuf */
 static void
-draw_pixbuf_tiled_aa (GdkPixbuf *pixbuf, GnomeCanvasBuf *buffer)
+draw_pixbuf_tiled_aa (GdkPixbuf *pixbuf, EelCanvasBuf *buffer)
 {
 	int x, y;
 	int start_x, start_y;
@@ -890,14 +1100,14 @@ draw_pixbuf_tiled_aa (GdkPixbuf *pixbuf, GnomeCanvasBuf *buffer)
 
 	for (y = start_y; y < buffer->rect.y1; y += tile_height) {
 		for (x = start_x; x < buffer->rect.x1; x += tile_width) {
-			eel_gnome_canvas_draw_pixbuf (buffer, pixbuf, x, y);
+			canvas_draw_pixbuf (buffer, pixbuf, x, y);
 		}
 	}
 }
 
 /* draw the background on the anti-aliased canvas */
 void
-eel_background_draw_aa (EelBackground *background, GnomeCanvasBuf *buffer)
+eel_background_draw_aa (EelBackground *background, EelCanvasBuf *buffer)
 {	
 	g_return_if_fail (EEL_IS_BACKGROUND (background));
 
@@ -915,10 +1125,10 @@ eel_background_draw_aa (EelBackground *background, GnomeCanvasBuf *buffer)
 	     buffer->rect.x1  > (background->details->image_rect_x + background->details->image_rect_width) ||
 	     buffer->rect.y1  > (background->details->image_rect_y + background->details->image_rect_height)) {
 		if (background->details->is_solid_color) {
-			eel_gnome_canvas_fill_rgb (buffer,
-							background->details->solid_color.red >> 8,
-							background->details->solid_color.green >> 8,
-							background->details->solid_color.blue >> 8);
+			fill_rgb (buffer,
+				  background->details->solid_color.red >> 8,
+				  background->details->solid_color.green >> 8,
+				  background->details->solid_color.blue >> 8);
 		} else {
 			fill_canvas_from_gradient_buffer (buffer, background);
 		}
@@ -938,10 +1148,10 @@ eel_background_draw_aa (EelBackground *background, GnomeCanvasBuf *buffer)
 			/* Since the image has already been scaled, all these cases
 			 * can be treated identically.
 			 */
-			eel_gnome_canvas_draw_pixbuf (buffer,
-							   background->details->image,
-							   background->details->image_rect_x,
-							   background->details->image_rect_y);
+			canvas_draw_pixbuf (buffer,
+					    background->details->image,
+					    background->details->image_rect_x,
+					    background->details->image_rect_y);
 			break;
 		}
 	}
@@ -952,7 +1162,7 @@ eel_background_draw_aa (EelBackground *background, GnomeCanvasBuf *buffer)
 
 void
 eel_background_draw_to_canvas (EelBackground *background,
-			       GnomeCanvasBuf *buffer,
+			       EelCanvasBuf *buffer,
 			       int entire_width,
 			       int entire_height)
 {
@@ -992,6 +1202,14 @@ eel_background_set_color_no_emit (EelBackground *background,
 	return TRUE;
 }
 
+/* Use style->base as the default color instead of bg */
+void
+eel_background_set_use_base (EelBackground *background,
+			     gboolean use_base)
+{
+	background->details->use_base = use_base;
+}
+
 void
 eel_background_set_color (EelBackground *background,
 			  const char *color)
@@ -1019,17 +1237,30 @@ eel_background_load_image_callback (GnomeVFSResult error,
 
 	/* Just ignore errors. */
 	if (pixbuf != NULL) {
+#if 0
+		EelBackgroundImagePlacement style;
+#endif
+
 		g_object_ref (pixbuf);
 		background->details->image = pixbuf;
 		background->details->image_width_unscaled = gdk_pixbuf_get_width (pixbuf);
 		background->details->image_height_unscaled = gdk_pixbuf_get_height (pixbuf);
+
+#if 0
+		style = EEL_BACKGROUND_TILED;
+		g_signal_emit (background, signals[DETERMINE_IMAGE_PLACEMENT], 0,
+			       gdk_pixbuf_get_width (pixbuf),
+			       gdk_pixbuf_get_height (pixbuf),
+			       &style);
+		eel_background_set_image_placement_no_emit (background, style);
+#endif
 	}
 
-	g_signal_emit (G_OBJECT (background), signals[IMAGE_LOADING_DONE], 0,
+	g_signal_emit (background, signals[IMAGE_LOADING_DONE], 0,
 		       pixbuf != NULL || background->details->image_uri == NULL);
 
 	if (background->details->emit_after_load) {
-		g_signal_emit (GTK_OBJECT (background), signals[APPEARANCE_CHANGED], 0);
+		g_signal_emit (background, signals[APPEARANCE_CHANGED], 0);
 	}
 }
 
@@ -1087,6 +1318,7 @@ eel_background_set_image_uri_helper (EelBackground *background,
 	g_return_val_if_fail (EEL_IS_BACKGROUND (background), FALSE);
 
 	if (eel_strcmp (background->details->image_uri, image_uri) == 0) {
+		background->details->emit_after_load = emit_appearance_change;
 		return FALSE;
 	}
 
@@ -1135,8 +1367,11 @@ set_image_and_color_image_loading_done_callback (EelBackground *background, gboo
 	/* We always emit , even if the color didn't change, because the image change
 	 * relies on us doing it here.
 	 */
-	g_signal_emit (G_OBJECT (background), signals[SETTINGS_CHANGED], 0);
-	g_signal_emit (G_OBJECT (background), signals[APPEARANCE_CHANGED], 0);
+	g_signal_emit (background, signals[SETTINGS_CHANGED], 0);
+
+	if (!background->details->emit_after_load) {
+		g_signal_emit (background, signals[APPEARANCE_CHANGED], 0);
+	}
 }
 
 /* Use this fn to set both the image and color and avoid flash. The color isn't
@@ -1229,24 +1464,32 @@ eel_background_reset (EelBackground *background)
 }
 
 static void
-draw_background_callback (GnomeCanvas *canvas, GdkDrawable *drawable,
+draw_background_callback (EelCanvas *canvas, 
 			  int x, int y, int width, int height)
 {
 	EelBackground *background;
 	GdkGCValuesMask value_mask;
+	GdkDrawable *drawable;
 	GdkGCValues gc_values;
 	GdkGC *gc;
+	gboolean bg_changed;
 
 	background = eel_get_widget_background (GTK_WIDGET (canvas));
 	if (background == NULL) {
 		return;
 	}
 
+	drawable = canvas->layout.bin_window;
 	
-	eel_background_ensure_realized (background, GTK_WIDGET (canvas)->window,
-					MAX (GTK_LAYOUT (canvas)->hadjustment->upper, GTK_WIDGET (canvas)->allocation.width),
-					MAX (GTK_LAYOUT (canvas)->vadjustment->upper, GTK_WIDGET (canvas)->allocation.height));
+	bg_changed = eel_background_ensure_realized (background, GTK_WIDGET (canvas)->window,
+						     MAX (GTK_LAYOUT (canvas)->hadjustment->upper, GTK_WIDGET (canvas)->allocation.width),
+						     MAX (GTK_LAYOUT (canvas)->vadjustment->upper, GTK_WIDGET (canvas)->allocation.height));
 
+	if (bg_changed) {
+		/* The background changed, issue a full redraw */
+		/* This is needed to handle gradients that grow on resize */
+		gtk_widget_queue_draw (GTK_WIDGET (canvas));
+	}
 
 	/* Create a new gc each time.
 	 * If this is a speed problem, we can create one and keep it around,
@@ -1255,8 +1498,8 @@ draw_background_callback (GnomeCanvas *canvas, GdkDrawable *drawable,
 	 */
 	if (background->details->background_pixmap != NULL) {
 		gc_values.tile = background->details->background_pixmap;
-		gc_values.ts_x_origin = -x;
-		gc_values.ts_y_origin = -y;
+		gc_values.ts_x_origin = 0;
+		gc_values.ts_y_origin = 0;
 		gc_values.fill = GDK_TILED;
 		value_mask = GDK_GC_FILL | GDK_GC_TILE |
 			GDK_GC_TS_X_ORIGIN | GDK_GC_TS_Y_ORIGIN;
@@ -1268,33 +1511,13 @@ draw_background_callback (GnomeCanvas *canvas, GdkDrawable *drawable,
 	}
 		
     	gc = gdk_gc_new_with_values (drawable, &gc_values, value_mask);
-	gdk_draw_rectangle (drawable, gc, TRUE, 0, 0, width, height);
+	gdk_draw_rectangle (drawable, gc, TRUE, x, y, width, height);
 	g_object_unref (gc);
 
 	/* We don't want the draw from the base class to happen. */
 	g_signal_stop_emission_by_name (canvas, "draw_background");
 }
 
-static void
-render_background_callback (GnomeCanvas *canvas, GnomeCanvasBuf *buffer)
-{
-	EelBackground *background;
-			
-	background = eel_get_widget_background (GTK_WIDGET (canvas));
-	if (background == NULL) {
-		return;
-	}
-
-	eel_background_pre_draw (background,
-				 GTK_WIDGET (canvas)->allocation.width,
-				 GTK_WIDGET (canvas)->allocation.height);
-
-	eel_background_draw_aa (background, buffer);
-	
-	/* We don't want the render from the base class to happen. */
-	g_signal_stop_emission_by_name (canvas, "render_background");
-}
-
 static void
 eel_background_set_up_widget (EelBackground *background, GtkWidget *widget)
 {
@@ -1336,19 +1559,15 @@ eel_background_set_up_widget (EelBackground *background, GtkWidget *widget)
 static void
 eel_background_set_up_canvas (GtkWidget *widget)
 {
-	if (!GNOME_IS_CANVAS (widget)) {
+	if (!EEL_IS_CANVAS (widget)) {
 		return;
 	}
 	if (g_object_get_data (G_OBJECT (widget), "eel_background_canvas_is_set_up") != NULL) {
 		return;
 	}
 
-	gnome_canvas_set_dither (GNOME_CANVAS (widget), GDK_RGB_DITHER_MAX);
-
 	g_signal_connect (widget, "draw_background",
 			  G_CALLBACK (draw_background_callback), NULL);
-	g_signal_connect (widget, "render_background",
-			  G_CALLBACK (render_background_callback), NULL);
 
 	g_object_set_data (G_OBJECT (widget), "eel_background_canvas_is_set_up", widget);
 }
@@ -1363,6 +1582,29 @@ eel_widget_background_changed (GtkWidget *widget, EelBackground *background)
 	gtk_widget_queue_draw (widget);
 }
 
+/* Callback used when the style of a widget changes.  We have to regenerate its
+ * EelBackgroundStyle so that it will match the chosen GTK+ theme.
+ */
+static void
+widget_style_set_cb (GtkWidget *widget, GtkStyle *previous_style, gpointer data)
+{
+	EelBackground *background;
+	
+	background = EEL_BACKGROUND (data);
+	
+	eel_widget_background_changed (widget, background);
+}
+
+static void
+widget_realize_cb (GtkWidget *widget, gpointer data)
+{
+	EelBackground *background;
+	
+	background = EEL_BACKGROUND (data);
+	
+	eel_background_set_up_widget (background, widget);
+}
+
 /* Gets the background attached to a widget.
 
    If the widget doesn't already have a EelBackground object,
@@ -1404,6 +1646,15 @@ eel_get_widget_background (GtkWidget *widget)
 				 G_CALLBACK (eel_widget_background_changed), widget, G_CONNECT_SWAPPED);
 	eel_widget_background_changed (widget, background);
 
+	g_signal_connect_object (widget, "style_set",
+				 G_CALLBACK (widget_style_set_cb),
+				 background,
+				 0);
+	g_signal_connect_object (widget, "realize",
+				 G_CALLBACK (widget_realize_cb),
+				 background,
+				 0);
+
 	return background;
 }
 
