@@ -32,6 +32,7 @@
 
 #include <gtk/gtkmain.h>
 #include <gtk/gtksignal.h>
+#include <gtk/gtkeditable.h>
 #include <gtk/gtkwindow.h>
 #include <gdk/gdkkeysyms.h>
 #include <gtk/gtkclipboard.h>
@@ -66,92 +67,122 @@ enum {
 
 static guint signals[LAST_SIGNAL] = { 0 };
 
-static void eel_editable_label_class_init        (EelEditableLabelClass    *klass);
-static void eel_editable_label_init              (EelEditableLabel         *label);
-static void eel_editable_label_set_property      (GObject          *object,
-						  guint             prop_id,
-						  const GValue     *value,
-						  GParamSpec       *pspec);
-static void eel_editable_label_get_property      (GObject          *object,
-						  guint             prop_id,
-						  GValue           *value,
-						  GParamSpec       *pspec);
-static void eel_editable_label_finalize          (GObject          *object);
-static void eel_editable_label_size_request      (GtkWidget        *widget,
-						  GtkRequisition   *requisition);
-static void eel_editable_label_size_allocate     (GtkWidget        *widget,
-						  GtkAllocation    *allocation);
-static void eel_editable_label_state_changed     (GtkWidget        *widget,
-						  GtkStateType      state);
-static void eel_editable_label_style_set         (GtkWidget        *widget,
-						  GtkStyle         *previous_style);
-static void eel_editable_label_direction_changed (GtkWidget        *widget,
-						  GtkTextDirection  previous_dir);
-static gint eel_editable_label_expose            (GtkWidget        *widget,
-						  GdkEventExpose   *event);
+static void     eel_editable_label_editable_init           (GtkEditableClass      *iface);
+static void     eel_editable_label_class_init              (EelEditableLabelClass *klass);
+static void     eel_editable_label_init                    (EelEditableLabel      *label);
+static void     eel_editable_label_set_property            (GObject               *object,
+							    guint                  prop_id,
+							    const GValue          *value,
+							    GParamSpec            *pspec);
+static void     eel_editable_label_get_property            (GObject               *object,
+							    guint                  prop_id,
+							    GValue                *value,
+							    GParamSpec            *pspec);
+static void     eel_editable_label_finalize                (GObject               *object);
+static void     eel_editable_label_size_request            (GtkWidget             *widget,
+							    GtkRequisition        *requisition);
+static void     eel_editable_label_size_allocate           (GtkWidget             *widget,
+							    GtkAllocation         *allocation);
+static void     eel_editable_label_state_changed           (GtkWidget             *widget,
+							    GtkStateType           state);
+static void     eel_editable_label_style_set               (GtkWidget             *widget,
+							    GtkStyle              *previous_style);
+static void     eel_editable_label_direction_changed       (GtkWidget             *widget,
+							    GtkTextDirection       previous_dir);
+static gint     eel_editable_label_expose                  (GtkWidget             *widget,
+							    GdkEventExpose        *event);
+static void     eel_editable_label_realize                 (GtkWidget             *widget);
+static void     eel_editable_label_unrealize               (GtkWidget             *widget);
+static void     eel_editable_label_map                     (GtkWidget             *widget);
+static void     eel_editable_label_unmap                   (GtkWidget             *widget);
+static gint     eel_editable_label_button_press            (GtkWidget             *widget,
+							    GdkEventButton        *event);
+static gint     eel_editable_label_button_release          (GtkWidget             *widget,
+							    GdkEventButton        *event);
+static gint     eel_editable_label_motion                  (GtkWidget             *widget,
+							    GdkEventMotion        *event);
+static gint     eel_editable_label_key_press               (GtkWidget             *widget,
+							    GdkEventKey           *event);
+static gint     eel_editable_label_key_release             (GtkWidget             *widget,
+							    GdkEventKey           *event);
+static gint     eel_editable_label_focus_in                (GtkWidget             *widget,
+							    GdkEventFocus         *event);
+static gint     eel_editable_label_focus_out               (GtkWidget             *widget,
+							    GdkEventFocus         *event);
+static void     eel_editable_label_commit_cb               (GtkIMContext          *context,
+							    const gchar           *str,
+							    EelEditableLabel      *label);
+static void     eel_editable_label_preedit_changed_cb      (GtkIMContext          *context,
+							    EelEditableLabel      *label);
+static gboolean eel_editable_label_retrieve_surrounding_cb (GtkIMContext          *context,
+							    EelEditableLabel      *label);
+static gboolean eel_editable_label_delete_surrounding_cb   (GtkIMContext          *slave,
+							    gint                   offset,
+							    gint                   n_chars,
+							    EelEditableLabel      *label);
+static void     eel_editable_label_clear_layout            (EelEditableLabel      *label);
+static void     eel_editable_label_recompute               (EelEditableLabel      *label);
+static void     eel_editable_label_ensure_layout           (EelEditableLabel      *label,
+							    gboolean               include_preedit);
+static void     eel_editable_label_select_region_index     (EelEditableLabel      *label,
+							    gint                   anchor_index,
+							    gint                   end_index);
+static gboolean eel_editable_label_focus                   (GtkWidget             *widget,
+							    GtkDirectionType       direction);
+static void     eel_editable_label_move_cursor             (EelEditableLabel      *label,
+							    GtkMovementStep        step,
+							    gint                   count,
+							    gboolean               extend_selection);
+static void     eel_editable_label_delete_from_cursor      (EelEditableLabel      *label,
+							    GtkDeleteType          type,
+							    gint                   count);
+static void     eel_editable_label_copy_clipboard          (EelEditableLabel      *label);
+static void     eel_editable_label_cut_clipboard           (EelEditableLabel      *label);
+static void     eel_editable_label_paste                   (EelEditableLabel      *label,
+							    GdkAtom                selection);
+static void     eel_editable_label_paste_clipboard         (EelEditableLabel      *label);
+static void     eel_editable_label_select_all              (EelEditableLabel      *label);
+static void     eel_editable_label_do_popup                (EelEditableLabel      *label,
+							    GdkEventButton        *event);
+static void     eel_editable_label_toggle_overwrite        (EelEditableLabel      *label);
+static gint     eel_editable_label_move_forward_word       (EelEditableLabel      *label,
+							    gint                   start);
+static gint     eel_editable_label_move_backward_word      (EelEditableLabel      *label,
+							    gint                   start);
+static void     eel_editable_label_reset_im_context        (EelEditableLabel      *label);
+static void     eel_editable_label_check_cursor_blink      (EelEditableLabel      *label);
+static void     eel_editable_label_pend_cursor_blink       (EelEditableLabel      *label);
 
-static void eel_editable_label_realize           (GtkWidget        *widget);
-static void eel_editable_label_unrealize         (GtkWidget        *widget);
-static void eel_editable_label_map               (GtkWidget        *widget);
-static void eel_editable_label_unmap             (GtkWidget        *widget);
-static gint eel_editable_label_button_press      (GtkWidget        *widget,
-						  GdkEventButton   *event);
-static gint eel_editable_label_button_release    (GtkWidget        *widget,
-						  GdkEventButton   *event);
-static gint eel_editable_label_motion            (GtkWidget        *widget,
-						  GdkEventMotion   *event);
-static gint eel_editable_label_key_press (GtkWidget   *widget,
-					  GdkEventKey *event);
-static gint eel_editable_label_key_release (GtkWidget   *widget,
-					    GdkEventKey *event);
-static gint eel_editable_label_focus_in (GtkWidget     *widget,
-					 GdkEventFocus *event);
-static gint eel_editable_label_focus_out (GtkWidget     *widget,
-					  GdkEventFocus *event);
+/* Editable implementation: */
+static void     editable_insert_text_emit     (GtkEditable *editable,
+					       const gchar *new_text,
+					       gint         new_text_length,
+					       gint        *position);
+static void     editable_delete_text_emit     (GtkEditable *editable,
+					       gint         start_pos,
+					       gint         end_pos);
+static void     editable_insert_text          (GtkEditable *editable,
+					       const gchar *new_text,
+					       gint         new_text_length,
+					       gint        *position);
+static void     editable_delete_text          (GtkEditable *editable,
+					       gint         start_pos,
+					       gint         end_pos);
+static gchar *  editable_get_chars            (GtkEditable *editable,
+					       gint         start_pos,
+					       gint         end_pos);
+static void     editable_set_selection_bounds (GtkEditable *editable,
+					       gint         start,
+					       gint         end);
+static gboolean editable_get_selection_bounds (GtkEditable *editable,
+					       gint        *start,
+					       gint        *end);
+static void     editable_real_set_position    (GtkEditable *editable,
+					       gint         position);
+static gint     editable_get_position         (GtkEditable *editable);
 
-static void     eel_editable_label_commit_cb  (GtkIMContext *context,
-					       const gchar  *str,
-					       EelEditableLabel *label);
-static void    eel_editable_label_preedit_changed_cb (GtkIMContext *context,
-						      EelEditableLabel  *label);
-static gboolean eel_editable_label_retrieve_surrounding_cb (GtkIMContext *context,
-							    EelEditableLabel  *label);
-static gboolean eel_editable_label_delete_surrounding_cb (GtkIMContext *slave,
-							  gint          offset,
-							  gint          n_chars,
-							  EelEditableLabel  *label);
 
-static void eel_editable_label_clear_layout        (EelEditableLabel *label);
-static void eel_editable_label_recompute           (EelEditableLabel *label);
-static void eel_editable_label_ensure_layout       (EelEditableLabel *label,
-						    gboolean        include_preedit);
-static void eel_editable_label_select_region_index (EelEditableLabel *label,
-						    gint      anchor_index,
-						    gint      end_index);
 
-static gboolean eel_editable_label_focus             (GtkWidget         *widget,
-						      GtkDirectionType   direction);
-
-static void eel_editable_label_move_cursor        (EelEditableLabel        *label,
-						   GtkMovementStep  step,
-						   gint             count,
-						   gboolean         extend_selection);
-static void eel_editable_label_delete_from_cursor (EelEditableLabel        *label,
-						   GtkDeleteType    type,
-						   gint             count);
-static void eel_editable_label_copy_clipboard     (EelEditableLabel        *label);
-static void eel_editable_label_select_all         (EelEditableLabel        *label);
-static void eel_editable_label_do_popup           (EelEditableLabel        *label,
-						   GdkEventButton  *event);
-static void eel_editable_label_toggle_overwrite   (EelEditableLabel        *label);
-
-static gint eel_editable_label_move_forward_word  (EelEditableLabel        *label,
-						   gint             start);
-static gint eel_editable_label_move_backward_word (EelEditableLabel        *label,
-						   gint             start);
-static void eel_editable_label_reset_im_context   (EelEditableLabel  *label);
-static void eel_editable_label_check_cursor_blink (EelEditableLabel *label);
-static void eel_editable_label_pend_cursor_blink  (EelEditableLabel *label);
 
 static GtkMiscClass *parent_class = NULL;
 
@@ -175,7 +206,18 @@ eel_editable_label_get_type (void)
 	(GInstanceInitFunc) eel_editable_label_init,
       };
 
+      static const GInterfaceInfo editable_info =
+      {
+	(GInterfaceInitFunc) eel_editable_label_editable_init,	/* interface_init */
+	NULL,							/* interface_finalize */
+	NULL							/* interface_data */
+      };
+
+      
       label_type = g_type_register_static (GTK_TYPE_MISC, "EelEditableLabel", &label_info, 0);
+      g_type_add_interface_static (label_type,
+				   GTK_TYPE_EDITABLE,
+				   &editable_info);
     }
   
   return label_type;
@@ -240,6 +282,8 @@ eel_editable_label_class_init (EelEditableLabelClass *class)
   class->move_cursor = eel_editable_label_move_cursor;
   class->delete_from_cursor = eel_editable_label_delete_from_cursor;
   class->copy_clipboard = eel_editable_label_copy_clipboard;
+  class->cut_clipboard = eel_editable_label_cut_clipboard;
+  class->paste_clipboard = eel_editable_label_paste_clipboard;
   class->toggle_overwrite = eel_editable_label_toggle_overwrite;
   
   signals[MOVE_CURSOR] = 
@@ -508,6 +552,21 @@ eel_editable_label_class_init (EelEditableLabelClass *class)
 				"toggle_overwrite", 0);
 }
 
+static void
+eel_editable_label_editable_init (GtkEditableClass *iface)
+{
+  iface->do_insert_text = editable_insert_text_emit;
+  iface->do_delete_text = editable_delete_text_emit;
+  iface->insert_text = editable_insert_text;
+  iface->delete_text = editable_delete_text;
+  iface->get_chars = editable_get_chars;
+  iface->set_selection_bounds = editable_set_selection_bounds;
+  iface->get_selection_bounds = editable_get_selection_bounds;
+  iface->set_position = editable_real_set_position;
+  iface->get_position = editable_get_position;
+}
+
+
 static void 
 eel_editable_label_set_property (GObject      *object,
 				 guint         prop_id,
@@ -579,6 +638,7 @@ eel_editable_label_init (EelEditableLabel *label)
 {
   label->jtype = GTK_JUSTIFY_LEFT;
   label->wrap = FALSE;
+  label->wrap_mode = PANGO_WRAP_WORD;
 
   label->layout = NULL;
   label->text = NULL;
@@ -793,6 +853,23 @@ eel_editable_label_set_line_wrap (EelEditableLabel *label,
     }
 }
 
+
+void
+eel_editable_label_set_line_wrap_mode (EelEditableLabel *label,
+				       PangoWrapMode     mode)
+{
+  g_return_if_fail (EEL_IS_EDITABLE_LABEL (label));
+  
+  if (label->wrap_mode != mode)
+    {
+      label->wrap_mode = mode;
+      
+      gtk_widget_queue_resize (GTK_WIDGET (label));
+    }
+  
+}
+
+
 /**
  * eel_editable_label_get_line_wrap:
  * @label: a #EelEditableLabel
@@ -1045,7 +1122,7 @@ eel_editable_label_ensure_layout (EelEditableLabel *label,
 		}
 	      pango_layout_set_width (label->layout, width);
 	    }
-	  pango_layout_set_wrap (label->layout, PANGO_WRAP_CHAR);
+	  pango_layout_set_wrap (label->layout, label->wrap_mode);
 	}
       else		/* !label->wrap */
 	pango_layout_set_width (label->layout, -1);
@@ -1607,6 +1684,15 @@ eel_editable_label_button_press (GtkWidget      *widget,
 	    eel_editable_label_select_region_index (label, index, index);
 	}
   
+      return TRUE;
+    }
+  else if (event->button == 2 && event->type == GDK_BUTTON_PRESS)
+    {
+      get_layout_index (label, event->x, event->y, &index);
+      
+      eel_editable_label_select_region_index (label, index, index);
+      eel_editable_label_paste (label, GDK_SELECTION_PRIMARY);
+      
       return TRUE;
     }
   else if (event->button == 3 && event->type == GDK_BUTTON_PRESS)
@@ -1731,6 +1817,8 @@ eel_editable_label_select_region_index (EelEditableLabel *label,
       label->selection_end == end_index)
     return;
 
+  eel_editable_label_reset_im_context (label);
+
   label->selection_anchor = anchor_index;
   label->selection_end = end_index;
 
@@ -2025,6 +2113,7 @@ eel_editable_label_delete_text (EelEditableLabel *label,
       gtk_widget_queue_resize (GTK_WIDGET (label));
       
       g_object_notify (G_OBJECT (label), "text");
+      g_signal_emit_by_name (GTK_EDITABLE (label), "changed");
     }
 }
 
@@ -2091,6 +2180,7 @@ eel_editable_label_insert_text (EelEditableLabel *label,
   gtk_widget_queue_resize (GTK_WIDGET (label));
 
   g_object_thaw_notify (G_OBJECT (label));
+  g_signal_emit_by_name (GTK_EDITABLE (label), "changed");
 }
 
 /* Used for im_commit_cb and inserting Unicode chars */
@@ -2161,11 +2251,16 @@ eel_editable_label_delete_surrounding_cb (GtkIMContext *slave,
 					  gint          n_chars,
 					  EelEditableLabel  *label)
 {
-#if 0
-  gtk_editable_delete_text (GTK_EDITABLE (label),
-			    label->selection_end + offset,
-			    label->selection_end + offset + n_chars);
-#endif
+  gint start_index, end_index;
+  char *text;
+
+  text = label->text + label->selection_anchor;
+  start_index = label->selection_anchor +
+    g_utf8_offset_to_pointer (text, offset) - text;
+  end_index = label->selection_anchor +
+    g_utf8_offset_to_pointer (text, offset + n_chars) - text;
+  
+  eel_editable_label_delete_text (label, start_index, end_index);
 
   return TRUE;
 }
@@ -2637,6 +2732,73 @@ eel_editable_label_copy_clipboard (EelEditableLabel *label)
     }
 }
 
+static void
+eel_editable_label_cut_clipboard (EelEditableLabel *label)
+{
+  if (label->text)
+    {
+      gint start, end;
+      gint len;
+      
+      start = MIN (label->selection_anchor,
+                   label->selection_end);
+      end = MAX (label->selection_anchor,
+                 label->selection_end);
+
+      len = strlen (label->text);
+
+      if (end > len)
+        end = len;
+
+      if (start > len)
+        start = len;
+
+      if (start != end)
+	{
+	  gtk_clipboard_set_text (gtk_clipboard_get (GDK_SELECTION_CLIPBOARD),
+				  label->text + start, end - start);
+	  eel_editable_label_delete_text (label, start, end);
+	}
+    }
+}
+
+static void
+paste_received (GtkClipboard *clipboard,
+		const gchar  *text,
+		gpointer      data)
+{
+  EelEditableLabel *label = EEL_EDITABLE_LABEL (data);
+      
+  if (text)
+    {
+      gint tmp_pos;
+      
+      if (label->selection_end != label->selection_anchor)
+	eel_editable_label_delete_selection (label);
+
+      tmp_pos = label->selection_anchor;
+      eel_editable_label_insert_text (label, text, strlen (text), &tmp_pos);
+      eel_editable_label_select_region_index (label, tmp_pos, tmp_pos);
+    }
+
+  g_object_unref (G_OBJECT (label));
+}
+
+static void
+eel_editable_label_paste (EelEditableLabel *label,
+			  GdkAtom   selection)
+{
+  g_object_ref (G_OBJECT (label));
+  gtk_clipboard_request_text (gtk_widget_get_clipboard (GTK_WIDGET (label), selection),
+			      paste_received, label);
+}
+
+static void
+eel_editable_label_paste_clipboard (EelEditableLabel *label)
+{
+  eel_editable_label_paste (label, GDK_NONE);
+}
+
 static void
 eel_editable_label_select_all (EelEditableLabel *label)
 {
@@ -2716,64 +2878,319 @@ eel_editable_label_toggle_overwrite (EelEditableLabel *label)
   label->overwrite_mode = !label->overwrite_mode;
 }
 
+typedef struct
+{
+  EelEditableLabel *label;
+  gint button;
+  guint time;
+} PopupInfo;
+
 static void
-eel_editable_label_do_popup (EelEditableLabel       *label,
-			     GdkEventButton *event)
+popup_targets_received (GtkClipboard     *clipboard,
+			GtkSelectionData *data,
+			gpointer          user_data)
 {
   GtkWidget *menuitem, *submenu;
   gboolean have_selection;
+  gboolean clipboard_contains_text;
+  PopupInfo *info;
+  EelEditableLabel *label;
 
-  if (label->popup_menu)
-    gtk_widget_destroy (label->popup_menu);
+  info = user_data;
+  label = info->label;
+
+  if (GTK_WIDGET_REALIZED (label))
+    {
+      if (label->popup_menu)
+	gtk_widget_destroy (label->popup_menu);
   
-  label->popup_menu = gtk_menu_new ();
+      label->popup_menu = gtk_menu_new ();
 
-  gtk_menu_attach_to_widget (GTK_MENU (label->popup_menu),
-                             GTK_WIDGET (label),
-                             popup_menu_detach);
+      gtk_menu_attach_to_widget (GTK_MENU (label->popup_menu),
+				 GTK_WIDGET (label),
+				 popup_menu_detach);
 
-  have_selection =
-    label->selection_anchor != label->selection_end;
-
-
-  append_action_signal (label, label->popup_menu, GTK_STOCK_CUT, "cut_clipboard",
-                        FALSE);
-  append_action_signal (label, label->popup_menu, GTK_STOCK_COPY, "copy_clipboard",
-                        have_selection);
-  append_action_signal (label, label->popup_menu, GTK_STOCK_PASTE, "paste_clipboard",
-                        FALSE);
+      have_selection =
+	label->selection_anchor != label->selection_end;
   
-  menuitem = gtk_menu_item_new_with_label (_("Select All"));
-  g_signal_connect_object (GTK_OBJECT (menuitem), "activate",
-			   GTK_SIGNAL_FUNC (eel_editable_label_select_all), label,
-			   0);
-  gtk_widget_show (menuitem);
-  gtk_menu_shell_append (GTK_MENU_SHELL (label->popup_menu), menuitem);
+      clipboard_contains_text = gtk_selection_data_targets_include_text (data);
 
-  menuitem = gtk_separator_menu_item_new ();
-  gtk_widget_show (menuitem);
-  gtk_menu_shell_append (GTK_MENU_SHELL (label->popup_menu), menuitem);
+      append_action_signal (label, label->popup_menu, GTK_STOCK_CUT, "cut_clipboard",
+			    have_selection);
+      append_action_signal (label, label->popup_menu, GTK_STOCK_COPY, "copy_clipboard",
+			    have_selection);
+      append_action_signal (label, label->popup_menu, GTK_STOCK_PASTE, "paste_clipboard",
+			    clipboard_contains_text);
+  
+      menuitem = gtk_menu_item_new_with_label (_("Select All"));
+      g_signal_connect_object (GTK_OBJECT (menuitem), "activate",
+			       GTK_SIGNAL_FUNC (eel_editable_label_select_all), label,
+			       G_CONNECT_SWAPPED);
+      gtk_widget_show (menuitem);
+      gtk_menu_shell_append (GTK_MENU_SHELL (label->popup_menu), menuitem);
+
+      menuitem = gtk_separator_menu_item_new ();
+      gtk_widget_show (menuitem);
+      gtk_menu_shell_append (GTK_MENU_SHELL (label->popup_menu), menuitem);
       
-  menuitem = gtk_menu_item_new_with_label (_("Input Methods"));
-  gtk_widget_show (menuitem);
-  submenu = gtk_menu_new ();
-  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), submenu);
+      menuitem = gtk_menu_item_new_with_label (_("Input Methods"));
+      gtk_widget_show (menuitem);
+      submenu = gtk_menu_new ();
+      gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), submenu);
+      
+      gtk_menu_shell_append (GTK_MENU_SHELL (label->popup_menu), menuitem);
   
-  gtk_menu_shell_append (GTK_MENU_SHELL (label->popup_menu), menuitem);
-  
-  gtk_im_multicontext_append_menuitems (GTK_IM_MULTICONTEXT (label->im_context),
-					GTK_MENU_SHELL (submenu));
+      gtk_im_multicontext_append_menuitems (GTK_IM_MULTICONTEXT (label->im_context),
+					    GTK_MENU_SHELL (submenu));
 
-  g_signal_emit (GTK_OBJECT (label),
-		 signals[POPULATE_POPUP], 0,
-		 label->popup_menu);
+      g_signal_emit (GTK_OBJECT (label),
+		     signals[POPULATE_POPUP], 0,
+		     label->popup_menu);
+
+      if (info->button)
+	gtk_menu_popup (GTK_MENU (label->popup_menu), NULL, NULL,
+			NULL, NULL,
+			info->button, info->time);
+      else
+	{
+	  gtk_menu_popup (GTK_MENU (label->popup_menu), NULL, NULL,
+			  popup_position_func, label,
+			  info->button, info->time);
+	  gtk_menu_shell_select_first (GTK_MENU_SHELL (label->popup_menu), FALSE);
+	}
+    }
+
+  g_object_unref (label);
+  g_free (info);
+}
+
+static void
+eel_editable_label_do_popup (EelEditableLabel *label,
+			     GdkEventButton   *event)
+{
+  PopupInfo *info = g_new (PopupInfo, 1);
+
+  /* In order to know what entries we should make sensitive, we
+   * ask for the current targets of the clipboard, and when
+   * we get them, then we actually pop up the menu.
+   */
+  info->label = g_object_ref (label);
   
   if (event)
-    gtk_menu_popup (GTK_MENU (label->popup_menu), NULL, NULL,
-                    NULL, NULL,
-                    event->button, event->time);
+    {
+      info->button = event->button;
+      info->time = event->time;
+    }
   else
-    gtk_menu_popup (GTK_MENU (label->popup_menu), NULL, NULL,
-                    popup_position_func, label,
-                    0, gtk_get_current_event_time ());
+    {
+      info->button = 0;
+      info->time = gtk_get_current_event_time ();
+    }
+
+  gtk_clipboard_request_contents (gtk_widget_get_clipboard (GTK_WIDGET (label), GDK_SELECTION_CLIPBOARD),
+				  gdk_atom_intern ("TARGETS", FALSE),
+				  popup_targets_received,
+				  info);
 }
+
+/************ Editable implementation ****************/
+
+static void
+editable_insert_text_emit  (GtkEditable *editable,
+			    const gchar *new_text,
+			    gint         new_text_length,
+			    gint        *position)
+{
+  EelEditableLabel *label = EEL_EDITABLE_LABEL (editable);
+  gchar buf[64];
+  gchar *text;
+  int text_length;
+
+  text_length = g_utf8_strlen (label->text, -1);
+
+  if (*position < 0 || *position > text_length)
+    *position = text_length;
+  
+  g_object_ref (G_OBJECT (editable));
+  
+  if (new_text_length <= 63)
+    text = buf;
+  else
+    text = g_new (gchar, new_text_length + 1);
+
+  text[new_text_length] = '\0';
+  strncpy (text, new_text, new_text_length);
+  
+  g_signal_emit_by_name (editable, "insert_text", text, new_text_length, position);
+
+  if (new_text_length > 63)
+    g_free (text);
+
+  g_object_unref (G_OBJECT (editable));
+}
+
+static void
+editable_delete_text_emit (GtkEditable *editable,
+			   gint         start_pos,
+			   gint         end_pos)
+{
+  EelEditableLabel *label = EEL_EDITABLE_LABEL (editable);
+  int text_length;
+
+  text_length = g_utf8_strlen (label->text, -1);
+
+  if (end_pos < 0 || end_pos > text_length)
+    end_pos = text_length;
+  if (start_pos < 0)
+    start_pos = 0;
+  if (start_pos > end_pos)
+    start_pos = end_pos;
+  
+  g_object_ref (G_OBJECT (editable));
+
+  g_signal_emit_by_name (editable, "delete_text", start_pos, end_pos);
+
+  g_object_unref (G_OBJECT (editable));
+}
+
+static void
+editable_insert_text (GtkEditable *editable,
+		      const gchar *new_text,
+		      gint         new_text_length,
+		      gint        *position)
+{
+  EelEditableLabel *label = EEL_EDITABLE_LABEL (editable);
+  gint index;
+
+  if (new_text_length < 0)
+    new_text_length = strlen (new_text);
+
+  index = g_utf8_offset_to_pointer (label->text, *position) - label->text;
+
+  eel_editable_label_insert_text (label,
+				  new_text,
+				  new_text_length,
+				  &index);
+  
+  *position = g_utf8_pointer_to_offset (label->text, label->text + index);
+}
+
+static void
+editable_delete_text (GtkEditable *editable,
+		      gint         start_pos,
+		      gint         end_pos)
+{
+  EelEditableLabel *label = EEL_EDITABLE_LABEL (editable);
+  int text_length;
+  gint start_index, end_index;
+
+  text_length = g_utf8_strlen (label->text, -1);
+
+  if (end_pos < 0 || end_pos > text_length)
+    end_pos = text_length;
+  if (start_pos < 0)
+    start_pos = 0;
+  if (start_pos > end_pos)
+    start_pos = end_pos;
+
+  start_index = g_utf8_offset_to_pointer (label->text, start_pos) - label->text;
+  end_index = g_utf8_offset_to_pointer (label->text, end_pos) - label->text;
+  
+  eel_editable_label_delete_text (label, start_index, end_index);
+}
+
+static gchar *    
+editable_get_chars (GtkEditable   *editable,
+		    gint           start_pos,
+		    gint           end_pos)
+{
+  EelEditableLabel *label = EEL_EDITABLE_LABEL (editable);
+  int text_length;
+  gint start_index, end_index;
+  
+  text_length = g_utf8_strlen (label->text, -1);
+
+  if (end_pos < 0 || end_pos > text_length)
+    end_pos = text_length;
+  if (start_pos < 0)
+    start_pos = 0;
+  if (start_pos > end_pos)
+    start_pos = end_pos;
+
+  start_index = g_utf8_offset_to_pointer (label->text, start_pos) - label->text;
+  end_index = g_utf8_offset_to_pointer (label->text, end_pos) - label->text;
+
+  return g_strndup (label->text + start_index, end_index - start_index);
+}
+
+static void
+editable_set_selection_bounds (GtkEditable *editable,
+			       gint         start,
+			       gint         end)
+{
+  EelEditableLabel *label = EEL_EDITABLE_LABEL (editable);
+  int text_length;
+  gint start_index, end_index;
+  
+  text_length = g_utf8_strlen (label->text, -1);
+
+  if (end < 0 || end > text_length)
+    end = text_length;
+  if (start < 0)
+    start = text_length;
+  if (start > text_length)
+    start = text_length;
+  
+  eel_editable_label_reset_im_context (label);
+
+  start_index = g_utf8_offset_to_pointer (label->text, start) - label->text;
+  end_index = g_utf8_offset_to_pointer (label->text, end) - label->text;
+
+  eel_editable_label_select_region_index (label, start_index, end_index);
+}
+
+static gboolean
+editable_get_selection_bounds (GtkEditable *editable,
+			       gint        *start,
+			       gint        *end)
+{
+  EelEditableLabel *label = EEL_EDITABLE_LABEL (editable);
+
+  *start = g_utf8_pointer_to_offset (label->text, label->text + label->selection_anchor);
+  *end = g_utf8_pointer_to_offset (label->text, label->text + label->selection_end);
+
+  return (label->selection_anchor != label->selection_end);
+}
+
+static void
+editable_real_set_position (GtkEditable *editable,
+			    gint         position)
+{
+  EelEditableLabel *label = EEL_EDITABLE_LABEL (editable);
+  int text_length;
+  int index;
+  
+  text_length = g_utf8_strlen (label->text, -1);
+  
+  if (position < 0 || position > text_length)
+    position = text_length;
+
+  index = g_utf8_offset_to_pointer (label->text, position) - label->text;
+  
+  if (index != label->selection_anchor ||
+      index != label->selection_end)
+    {
+      eel_editable_label_select_region_index (label, index, index);
+    }
+}
+
+static gint
+editable_get_position (GtkEditable *editable)
+{
+  EelEditableLabel *label = EEL_EDITABLE_LABEL (editable);
+  
+  return g_utf8_pointer_to_offset (label->text, label->text + label->selection_anchor);
+}
+
+
+     
